// # Filename: fragmentShader3D.glsl
// #
// # Synopsis: This shader carries out ray-casting based 3D rendition.
// #           Three textures are necessary: 3D volume data, 3D
// # 		control data and 1D transfer function.
// #
// # Copyright (C) 2010 
// #	 Author: Wu Shin-Ting (26/12/2010), Lionis de Souza Watanabe and 
// #	         Jose Elias Yauri Vidalon
// #
// #   This program is free software: you can redistribute it and/or modify
// #   it under the terms of the GNU Affero General Public License as
// #   published by the Free Software Foundation, either version 3 of the
// #   License, or (at your option) any later version.
// #
// #   This program is distributed in the hope that it will be useful,
// #   but WITHOUT ANY WARRANTY; without even the implied warranty of
// #   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// #   GNU Affero General Public License for more details.
// #
// #   You should have received a copy of the GNU Affero General Public License
// #   along with this program.  If not, see <http://www.gnu.org/licenses/>.
// #

uniform sampler3D IMAGE3D;
uniform sampler3D CONTROLVOLUME;
uniform sampler1D TRANSFERFUNCTION1D;
uniform vec4 scaleFactor;
uniform vec3 extents;
uniform vec3 cursorPosition;
uniform bvec3 drawPlaneSide1;
uniform bvec3 drawPlaneSide2;
uniform bvec3 clippingPlaneSide1;
uniform bvec3 clippingPlaneSide2;
uniform float stepSize;
uniform float depth_slicing;
uniform float noise_threshold;
uniform bool flipPlaneSide1;
uniform vec4 cutPlanePoint; //(x,y,z,1.0)
uniform vec4 cutPlaneVec; //(x,y,z,0.0)

float z_value;

// Find the ray intersection interval with respect 
// to the proxy geometry  (Liang Barsky)
bool intersectBox(in vec4 r_o, in vec4 r_d, in vec4 boxmin, in vec4 boxmax, 
inout float tnear, inout float tfar)
{
    // compute intersection of ray with all six bbox planes (Liang-Barsky)
    vec4 invR = vec4(1.0,1.0,1.0,1.0)/r_d;
    vec4 tbot = invR * (boxmin - r_o);
    vec4 ttop = invR * (boxmax - r_o);

    // re-order intersections to find smallest and largest on each axis
    vec4 tmin = min(ttop, tbot);
    vec4 tmax = max(ttop, tbot);

    if (all(notEqual (tmin.xyz,tmax.xyz))) {
        // find the largest tmin and the smallest tmax
    	float largest_tmin = max(max(tmin.x, tmin.y), max(tmin.x, tmin.z));
    	float smallest_tmax = min(min(tmax.x, tmax.y), min(tmax.x, tmax.z));

    	tnear = largest_tmin;
    	tfar = smallest_tmax;

    	if (tfar >= tnear) return true;
    	else return false;
    } else {
      // r_d is parallel to at least two planes. Just consider the range [0,1].
      // It should be improved.
      tnear = 0.;
      tfar = 1.0;
      return true;
    }
}

//Process geometric position from any voxel,
//copying its depth to depth buffer.
//thus any voxels between a observer and this voxel
//will be ignored from visualization
void processZvalue(in vec4 geomPos, inout bool data_found){
	vec4 centralizedGeomPos, boundary;
	mat4 mvp;
	
	mvp = gl_ModelViewProjectionMatrix;
	boundary.xyz = extents.xyz;
	boundary.w = 1.0;

	centralizedGeomPos = boundary * geomPos; // geomPos range: [0,1]; centralizedGeomPos range: [-1,1]
	z_value = (mvp * centralizedGeomPos).z;// transform the centralized geomPos
	z_value = 0.5 * z_value + 0.5;//put it in range [0,1]
	
	data_found = true;
}

//Process all control tags (markups and depths, at this moment)
void processControlTags(in float image3Dscalar, in float TFscalar, 
in vec4 geomPos, inout bool data_found, inout vec4 dstColor){
	int markup;
	float depth;
	vec4 controlTag;

	controlTag = texture3D(CONTROLVOLUME, geomPos.xyz);
	markup = int(controlTag.r * float(INT_FORMAT_BITS));
	
	//all markups (except transparency) are considered valid voxels
	//to be included in depth buffer
	switch (markup){
	case TAG001://transparent
		break;
	case TAG002://painted
		dstColor = vec4(0.0, 1.0, 1.0, 1.0);
		processZvalue(geomPos, data_found);
		break;
	case TAG003://point of reference
		dstColor = vec4(1.0, 0.0, 1.0, 1.0);
		processZvalue(geomPos, data_found);
		break;
	case TAG004://point of interest
		dstColor = vec4(1.0, 1.0, 0.0, 1.0);
		processZvalue(geomPos, data_found);
		break;
	case TAG005://line
		dstColor = vec4(0.0, 1.0, 0.0, 1.0);
		processZvalue(geomPos, data_found);
		break;
	case TAG006://eraser
		dstColor = vec4(1.0, 1.0, 0.0, 1.0);
		processZvalue(geomPos, data_found);
		break;		
	/*case TAG007://selection voxel
		dstColor = vec4(0.0, 1.0, 1.0, 0.2);
		processZvalue(geomPos, data_found);
		break;*/
		
	default: //if there wasn't a markup tag, begins depth processing 
		depth = controlTag.y;//get voxel's depth //Revised by Ting
		
		//zvalue is set once, only to the first valid voxel in ray composition.
		//so, if some valid voxel was already found, depth buffer from previous voxel
		//won't be ovewritten, ignoring this step
		if(!data_found)
		
			//1) all untagged voxels have depth=0 and must be shown.
			//2) also, voxels deeper (with greater depth) than the slicing mesh have to be shown.
			if(depth == 0. || depth_slicing <= depth)
			
				//a scalar greater than noise threshold determines a valid voxel
				if(image3Dscalar > noise_threshold)
				
					//TFscalar is what the user is seeing indeed,
					//so consider only valid entries from transfer function
					if(TFscalar > 0.0)
						processZvalue(geomPos, data_found);
		break;
	}
}

//Process the axis aligned planes, showing them and
//doing volume clipping if desired by the user
void processAxisPlanes(in bool flipPlane, in vec4 geomPos, in vec4 axialColor, 
in vec4 sagittalColor, in vec4 coronalColor, inout bool data_found, 
inout vec4 dstColor){

	//Drawing planes from side 1
	if (drawPlaneSide1[VMTK_AXIAL] == true && geomPos.y < cursorPosition.z + stepSize 
			&& geomPos.y > cursorPosition.z ) {
		dstColor += axialColor;
		processZvalue(geomPos, data_found);
	}
	if (drawPlaneSide1[VMTK_SAGITTAL] == true && geomPos.z < cursorPosition.x + stepSize
			&& geomPos.z > cursorPosition.x ) {
		dstColor += sagittalColor;
		processZvalue(geomPos, data_found);
	}
	if (drawPlaneSide1[VMTK_CORONAL] == true && geomPos.x < cursorPosition.y + stepSize
			&& geomPos.x > cursorPosition.y ) {
		dstColor += coronalColor;
		processZvalue(geomPos, data_found);
	}

	//Clipping planes from side 1
	if (!flipPlane){
		if ((clippingPlaneSide1[VMTK_AXIAL] == true && geomPos.y > cursorPosition.z  + stepSize )
			|| (clippingPlaneSide1[VMTK_SAGITTAL] == true && geomPos.z > cursorPosition.x + stepSize )
			|| (clippingPlaneSide1[VMTK_CORONAL] == true && geomPos.x > cursorPosition.y + stepSize )){

			//no color until ray is passing through clipped portion of volume
			dstColor = vec4(0.0);
			z_value = 1.0;

			//the clipped portion of volume isn't valid data to be 
			//included in ray composition, so no data found until this position
			data_found = false;
		}
	}else {
		if ((clippingPlaneSide1[VMTK_AXIAL] == true && geomPos.y < cursorPosition.z  + stepSize )
			|| (clippingPlaneSide1[VMTK_SAGITTAL] == true && geomPos.z < cursorPosition.x + stepSize )
			|| (clippingPlaneSide1[VMTK_CORONAL] == true && geomPos.x < cursorPosition.y + stepSize )){

			//no color until ray is passing through clipped portion of volume
			dstColor = vec4(0.0);
			z_value = 1.0;

			//the clipped portion of volume isn't valid data to be 
			//included in ray composition, so no data found until this position
			data_found = false;
		}
	}//EndIf FlipPlane
		
}

void main(void) {

	bool data_found, hit;
	float scalar, t, tnear, tfar;
	vec4 geomPos, initGeomPos, volExtentMin, volExtentMax;
	vec4 axialColor, sagittalColor, coronalColor;
	vec4 dstColor, srcColor, srcColor1, srcColor2, camDir;
	vec4 geomDir;
	mat4 mvi, mv;

	axialColor = vec4(0.0, 0.08, 0.08, 0.2);//color of axial plane
	sagittalColor = vec4(0.08, 0.0, 0.08, 0.2);//color of sagittal plane
	coronalColor = vec4(0.08, 0.08, 0.0, 0.2);//color of coronal plane

	volExtentMin = vec4(0.0, 0.0, 0.0, 0.0);//min value of boundingbox
	volExtentMax = vec4(1.0, 1.0, 1.0, 1.0);//max value of boundingbox

	dstColor = srcColor = srcColor1= srcColor2= vec4(0.0);//initialize color
	z_value = 1.0;

	data_found = false; //flag used when valid voxels are found
	camDir = vec4(0.0, 0.0, -1.0, 0.0);//camera direction, used for ray composition

	mvi = gl_ModelViewMatrixInverse;//inverse model view matrix
	mv = gl_ModelViewMatrix;//inverse model view matrix

	//determine volume entry position / ray's starting point
	initGeomPos = gl_TexCoord[1];
	initGeomPos.w = 1.0;

	//compute ray direction
	geomDir = normalize(mvi * camDir) * scaleFactor;
	geomDir.w = 0.0;

	hit = intersectBox(initGeomPos, geomDir, volExtentMin, volExtentMax, tnear, tfar);
    if (!hit) {
       gl_FragColor = vec4(1.0,0.0,0.0,1.0);       
       return;
    }        

    // Get the normal vector of cutting plane
    // It is necessary to apply scale factor due to texture mapping
    vec4 pVec, pt;
    if (cutPlaneVec.x == 0.0 && cutPlaneVec.y == 0.0 && cutPlaneVec.z == 0.0)
       	pVec = vec4(0.,0.,0.,0.);
    else {
        vec4 inv = vec4(1./scaleFactor.x,1./scaleFactor.y,1./scaleFactor.z,1.0);
		pVec = normalize(cutPlaneVec*inv);
    }
    // Determine the offset of the cutting plane
    pt = cutPlanePoint*scaleFactor;
    float d = -(pVec.x*pt.x + pVec.y*pt.y + pVec.z*pt.z);

    if (tnear < 0.0) tnear = 0.0;     // clamp to near plane    
	t = tnear;
	
	///the ray traversal loop checks if the ray is still inside of volume extents
	//and if opacity haven't reached maximum value (opacity = 1) 
	while (t < tfar && dstColor.a < 0.99) {

	  geomPos = initGeomPos + geomDir * t;

      if (geomPos.x*pVec.x + geomPos.y*pVec.y + geomPos.z*pVec.z + d <= 0.0) {
		scalar = texture3D(IMAGE3D, geomPos.xyz).r;//get RAW scalar value
		srcColor = texture1D(TRANSFERFUNCTION1D, scalar);//apply transfer function
		srcColor.rgb *= srcColor.a; //compute associated color of voxel				
			
		//control volume processing determines depth buffer and
		//the beginning of ray composition
		processControlTags(scalar, srcColor.a, geomPos, data_found, dstColor);
		// Draw or clip the volume
		processAxisPlanes(flipPlaneSide1, geomPos, axialColor, sagittalColor, 
					coronalColor, data_found, dstColor);
			
		// If a valid voxel was found, make front-to-back ray composition
		if(data_found){									
			dstColor.rgb += mix(vec3(0.0), srcColor.rgb, 1.0 - dstColor.a);			
			dstColor.a += (1.0 - dstColor.a) * srcColor.a;				
		}		
                }			
		//advance ray position along ray direction
		t += stepSize;
	  }
		
    if (dstColor.a == 0.0) 
		gl_FragColor = vec4(0.5, 0.5, 0.5, 1.0);
	else  
		gl_FragColor = dstColor;
}
